/*
        Text.c 1.0 for X11R6 & GNU C Compiler

        Copyright (C) 1998 Syllabub
        Maroontress Fast Software.
*/

#include <stdlib.h>
#include <stdio.h>

#include "WCString.h"
#include "Text.h"

SylText *
CreateSylTextFromWCString(wchar_t *wcs, int eol)
{
    SylText *txt;
    
    if ((txt = (SylText *)malloc(sizeof(SylText))) == NULL)
	return (NULL);
    txt->len = wstrlen(wcs);
    txt->eol = eol;
    if ((txt->wcs = (wchar_t *)malloc((txt->len + 1) * sizeof(wchar_t)))
	== NULL) {
	free(txt);
	return (NULL);
    }
    wstrcpy(txt->wcs, wcs);
    return (txt);
}

void
FreeSylText(SylText *txt)
{
    free(txt->wcs);
    free(txt);
}

wchar_t *
CreateWCStringFromSylText(SylText *txt, int bgn, int end)
{
    wchar_t *wcs;

    if ((wcs = (wchar_t *)malloc((end - bgn + 1) * sizeof(wchar_t))) == NULL)
	return (NULL);
    wstrncpy(wcs, txt->wcs + bgn, end - bgn);
    wcs[end - bgn] = 0;
    return (wcs);
}

void
InsertWCharIntoSylText(SylText *txt, int n, wchar_t c)
{
    wchar_t *sum;
    int len;

    len = txt->len + 1;
    if ((sum = (wchar_t *)malloc((len + 1) * sizeof(wchar_t))) == NULL)
	return;
    wstrncpy(sum, txt->wcs, n);
    sum[n] = c;
    wstrcpy(sum + n + 1, txt->wcs + n);
    free(txt->wcs);
    txt->wcs = sum;
    txt->len = len;
}

void
InsertWCStringIntoSylText(SylText *txt, int n, wchar_t *p)
{
    while (*p) {
	InsertWCharIntoSylText(txt, n, *p);
	++n;
	++p;
    }
}

void
DeleteWCharOfSylText(SylText *txt, int n)
{
    wchar_t *sum;
    int len;

    if (txt->len <= 0 || txt->len <= n)
	return;
    len = txt->len - 1;
    if ((sum = (wchar_t *)malloc((len + 1) * sizeof(wchar_t))) == NULL)
	return;
    wstrncpy(sum, txt->wcs, n);
    if (txt->wcs[n] != 0)
	wstrcpy(sum + n, txt->wcs + n + 1);
    free(txt->wcs);
    txt->wcs = sum;
    txt->len = len;
}

void
DeleteWCStringOfSylText(SylText *txt, int n, int m)
{
    for (m -= n; m > 0; --m) {
	DeleteWCharOfSylText(txt, n);
    }
}

void
ChangeEndOfLineSylText(SylText *txt, int eol)
{
    txt->eol = eol;
}

/*
  head$B$N8e$K(Btail$B$rO"7k$7$?FbMF$HEy$7$$(BSylText$B$r:n@.$7JV$9(B.
*/
SylText *
ConcatAndCreateSylText(SylText *head, SylText *tail)
{
    wchar_t *s, *d;
    SylText *txt;
    
    if ((txt = (SylText *)malloc(sizeof(SylText))) == NULL)
	return (NULL);
    txt->len = LengthOfSylText(head) + LengthOfSylText(tail);
    txt->eol = EndOfLineSylText(tail);
    if ((txt->wcs = (wchar_t *)malloc((txt->len + 1) * sizeof(wchar_t)))
	== NULL) {
	free(txt);
	return (NULL);
    }
    d = txt->wcs;
    for (s = head->wcs; *s != 0; ++s)
	*d++ = *s;
    for (s = tail->wcs; *s != 0; ++s)
	*d++ = *s;
    *d = 0;
    return (txt);
}

/*
  txt$B$N(Bn$BJ8;zL\$^$G$NFbMF$r(Bhead$B$K!"(Bn$BJ8;zL\0J9_$NFbMF$r(Btail$B$K:n@.$9$k(B.
*/
void
SplitAndCreateSylTexts(SylText *txt, int n, int eol,
		       SylText **head, SylText **tail)
{
    if ((*head = (SylText *)malloc(sizeof(SylText))) != NULL) {
	(*head)->wcs = CreateWCStringFromSylText(txt, 0, n);
	(*head)->len = n;
	(*head)->eol = eol;
    }
    if ((*tail = (SylText *)malloc(sizeof(SylText))) != NULL) {
	(*tail)->wcs = CreateWCStringFromSylText(txt, n, LengthOfSylText(txt));
	(*tail)->len = LengthOfSylText(txt) - n;
	(*tail)->eol = EndOfLineSylText(txt);
    }
}
